/*
VocaTest is intended as a small and simple tool for testing vocabularies
Copyright (C) 2003, 2004  Olof Holmgren

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// 3 includes for readline support
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <iostream>
#include <fstream>
#include <string>
#include <deque>
#include "StringHandling.h" 
#include "VocPair.h"
#include "VocList.h"
#include "FileList.h"
#include "Strings.h"
using namespace std;


// Declaring some functions
void programLoop();
void setGlobals(const string &fileName);
void dispHelp();
void load(string &file, const bool append);
void listVocs();
void listFiles();
void test(const bool backw, const bool strict, const bool failedOnly);

bool isAnsInList(deque<string>& ansList, const string& ans);
string testVoc(VocPair* vp, const bool& backw,
	       const unsigned int& vocNr, const unsigned int& nrOfVocs);
void save(string &file);
void randomize();
void buildVoc(string &file);
void unappend(string &file);

void testStrict(VocPair* vp, const bool& backw, bool& abort,
		const unsigned int& vocNr, const unsigned int& nrOfVocs);
bool testNormal(VocPair* vp, const bool& backw, bool& abort,
		const unsigned int& vocNr, const unsigned int& nrOfVocs);
bool ifSettingApply(string &row);
bool wantsToOverwriteFile(const std::string& file);
void writeIfPossible(string& file, const string& data);
void fullPathify(string& file);

// Located in Strings.h
extern string helpString;
extern string sampleFile;
extern string startupString;


// Declaring da global var's...
VocList vocList;
FileList fileList;
string setting_langDelim = "|";
string setting_wordDelim = "";
int setting_nrOfRetrys = 1;
string setting_vocDir = "~/";
bool setting_useVocDir = false;

// Da main thing
int main(int argc, char* argv[])
{
  if((argc > 1) && (((string)argv[1] == "--version") || (string)argv[1] == "-v"))
    {
      cout << "VocaTest version 2.0" << endl;
      return 0;
    }
  else if((argc > 1) && (((string)argv[1] == "--help") || (string)argv[1] == "-h"))
    {
      cout <<
	"Commandline options:\n"
	"--help, -h \t\t Show this text \n"
	"--test, -t [files] \t Start a test on the given files \n"
	"--version, -v \t\t Display current version \n";
      return 0;
    } 
  else if((argc > 1) && (((string)argv[1] == "--test") || (string)argv[1] == "-t"))
    {
      for(int i = 2; i < argc; i++)
	{
	  string file = argv[i];
	  load(file, true);
	}
      test(false, false, false);
      return 0;
    }

  // SETUP
  string settingFile = "~/.VocaTest";
  fullPathify(settingFile);

  ifstream inFile(settingFile.c_str());
  bool fileExists = inFile; // True if exists, false otherwise
  if(!fileExists)
    {
      cout << endl;
      cout << "=======================================================" << endl;
      cout << "Setting up VocaTest..." << endl;
      cout << "Creating file " << settingFile << endl;
      writeIfPossible(settingFile, settingString);      
      cout << "Setup done" << endl;
      cout << "=======================================================" << endl;
    }
  // End of SETUP
  
  setGlobals(settingFile);
  programLoop();

  return 0;
}
  
// Go through a given file for settings and apply these
void setGlobals(const string &file)
{
  ifstream inFile(file.c_str());

  if(!inFile)
    {
      cerr << "Could not open " << file << " for reading. Using default settings."
	   << endl;
      return;
    }
  
  string row;
  while(getline(inFile, row))
    ifSettingApply(row);

  inFile.close();
  
  return;
}

// The main loop thing
void programLoop()
{
  cout << endl << startupString << endl;
  dispHelp();
  
  while(true)
    {      
      string input = readline("<VocaTest>");
      if(input != "") // don't save 'nothings' in the history
	add_history(input.c_str());

      string command = "";
      string arg = "";
      
      // Filter out the actual command and remove it from the input string
      rmLeadWs(input);
      rmTrailWs(input);
      if(input.find(' ') != string::npos)
	{
	  int spacePos = input.find(' ');
	  
	  command = input.substr(0, spacePos);
	  arg = input.substr(spacePos, input.length());
	}
      else // No space found ==> no args!
	command = input;
      
      // Filter out the argument, if any
      rmLeadWs(arg);
      rmTrailWs(arg);
      
      // Execute the command!
      if(command == "load")
	load(arg, false);
      else if(command == "append")
	load(arg, true);
      else if(command == "unappend")
	unappend(arg);
      else if(command == "list")
	listVocs();
      else if(command == "listfiles")
	listFiles();
      else if(command == "test" || command == "testbackw")
	{
	  bool backw = false;
	  bool strict = false;
	  bool failed = false;

	  if(command == "testbackw")
	    backw = true;
	  // else 'test' and backw = false already
	  
	  deque<string> args = split(arg, " ");
	  for(unsigned int i = 0; i < args.size(); i++)
	    { 
	      if(args[i] == "strict")
		strict = true;
	      else if(args[i] == "failed")
		failed = true;
	    }
	  test(backw, strict, failed);
	}
      else if(command == "randomize")
	randomize();
      else if(command == "buildvoc")
	buildVoc(arg);
      else if(command == "save")
	save(arg);
      else if(command == "help")
	dispHelp();
      else if(command == "$abort")
	cout << "That command can only be executed while running a test" << endl;
      else if(command == "exit" || command == "quit" || command == "q")
	return;
      else
	cout << "No such command. Type 'help' to see the available commands" << endl;
    }
}

// Print the help
void dispHelp()
{
  cout << helpString << endl << endl;;
}

// Randomize the list
void randomize()
{
  vocList.randomize();
  cout << "All tests will now be performed in random order" << endl;
}

// Build the sample vocfile
void buildVoc(string& file)
{
  if(file == "")
    cout << "And under what filename should I build this vocfile?" << endl;
  else
    writeIfPossible(file, sampleFile);
}


// Build the sample vocfile
void save(string& file)
{
  if(vocList.size() == 0)
    cout << "No vocs loaded. I don't save things that do not exist" << endl;
  else if(file == "")
    cout << "And under what filename should I save this file?" << endl;
  else
    writeIfPossible(file, vocList.toString(setting_wordDelim, setting_langDelim));
}

// Load a file
void load(string &file, const bool append)
{
  if(file == "")
    {
      cout << "And what file am I supposed to read from?" << endl;
      return;
    }
  
  fullPathify(file);
  ifstream inFile(file.c_str());
  
  if(!inFile)
    {
      cout << file << " can't be opened for reading" << endl;
      inFile.close();
      return;
    }
  
  if(!append)
    {
      vocList.clear();
      fileList.clear();
    }
  
  string row;
  bool vocFound = false;
  while(getline(inFile, row))
    {
      if(!ifSettingApply(row))
	{
	  unsigned int pos = row.find(setting_langDelim);
	  // See to it that it's a valid voc
	  if((row.find(setting_langDelim) != string::npos) &&
	     (pos != row.length() - 1) && (pos != 0))
	    {
	      // Filter lang1 & lang2
	      string lang1W = row.substr(0, pos);
	      string lang2W = row.substr(pos + setting_langDelim.length(),
					 row.length());
	      
	      deque<string> lang1L = split(lang1W, setting_wordDelim);   
	      deque<string> lang2L = split(lang2W, setting_wordDelim);
	      unsigned int id = fileList.getNextId();
	      
	      vocList.addVoc(new VocPair(lang1L, lang2L, id));
	      vocFound = true;
	    }
	}
    }
  inFile.close();
  
  if(!vocFound)
    cout << file << " contains no valid vocs" << endl;
  else
    {
      fileList.addFile(file);
      cout << "Loaded file: " << file << endl;
    }
}

// Unappend a vocfile
void unappend(string &file)
{
  if(file == "")
    {
      cout << "And what file am I supposed to unappend?" << endl;
      return;
    }
  
  fullPathify(file);
  unsigned int idToRemove = 0;
  bool fileFound = false;
  for(unsigned int i = 0; i < fileList.size(); i++)
    if(fileList.getFileName(i) == file)
      {
	idToRemove = fileList.getId(i);
	fileList.remove(i);
	fileFound = true;
	break;
      }
  
  if(!fileFound)
    {
      cout << "File " << file << " is not loaded" << endl;
      return;
    }
  
  unsigned int i = 0;
  while(i < vocList.size())
    {
      // remove will change the vocList structure so that i++ wont be needed
      // but we have to call i++ when we didn't get a match
      if(vocList.getVoc(i)->getId() == idToRemove)
	vocList.remove(i); 
      else	
	i++;
    }
  cout << "File " << file << " unappended" << endl;
}

// Print the vocs
void listVocs()
{
  if(vocList.size() == 0)
    cout << "No voc(s) loaded" << endl;
  else
    cout << endl << vocList << endl
	 << "Number of vocs: " << vocList.size() << endl;
}

// List the laoded vocfiles
void listFiles()
{
  if(fileList.size() == 0)
    cout << "No file(s) loaded" << endl;
  else
    cout << endl << fileList << endl
	 << "Number of different files: " << fileList.size() << endl;
}

// Time to do a voctest...
void test(const bool backw, const bool strict, const bool failedOnly)
{
  if(vocList.size() == 0)
    {
      cout << "You have to load a vocfile first!" << endl;
      return;
    }

  if(vocList.isRandom())
    vocList.randomize();

  unsigned int nrOfVocs = 0;
  if(failedOnly)
    for(unsigned int i = 0; i < vocList.size(); i++)
      {
	if(!vocList.getVoc(i)->getPassed())
	  ++nrOfVocs;
      }
  else
    nrOfVocs = vocList.size();
  
  int passedWords = 0;
  bool abort = false;
  unsigned int i = 0;
  unsigned int vocNr = 1;
  while(!abort && (i < vocList.size()))
    {
      if(failedOnly && vocList.getVoc(i)->getPassed())
	{
	  // We are testing for failed words only and this
	  // voc was a pass last time around, so skip it!
	  // i++ so we won't end up at the same place at the next lap
	  ++i;
	  continue;
	}
      
      if(strict)
	testStrict(vocList.getVoc(i), backw, abort, vocNr, nrOfVocs);
      else
	if(testNormal(vocList.getVoc(i), backw, abort, vocNr, nrOfVocs))
	  ++passedWords;
      
      ++i;
      ++vocNr;
    }

  if(abort)
    cout << "Test aborted" << endl;
  else
    cout << "Nr of correct: " << passedWords << endl
	 << "Nr of failed words: " << nrOfVocs - passedWords << endl
	 << "Total number of words: " << nrOfVocs << endl;
}


////////////////////////////////////////
// Helper functions below
////////////////////////////////////////

// A strict test
void testStrict(VocPair* vp, const bool& backw, bool& abort,
		const unsigned int& vocNr, const unsigned int& nrOfVocs)
{
  deque<string> ansList = vp->getAnswers(backw);
  
  while(ansList.size() > 0)
    {
      string answer = testVoc(vp, backw, vocNr, nrOfVocs);
      if(answer == "$abort")
	{
	  abort = true;
	  break;
	}
      else if(isAnsInList(ansList, answer))
	cout << "Correct!" << endl << endl;
      else
	cout << "Wrong!" << endl
	     << "The correct answer was: "
	     << vp->getStringAnswers(backw)
	     << endl << endl;
    }
  return;
}

// A normal test
bool testNormal(VocPair* vp, const bool& backw, bool& abort,
		const unsigned int& vocNr, const unsigned int& nrOfVocs)
{
  deque<string> ansList = vp->getAnswers(backw);
  bool retval = false;
  for(int j = 0; j <= setting_nrOfRetrys; j++)
    {
      string answer = testVoc(vp, backw, vocNr, nrOfVocs);
      if(answer == "$abort")
	{
	  abort = true;
	  break;
	}
      else if(isAnsInList(ansList, answer))
	{
	  vp->setPassed(true);
	  cout << "Correct!" << endl << endl;
	  retval = true;
	  break;
	}
      else // Wrong
	{
	  if((setting_nrOfRetrys - j) == 0)
	    {
	      vp->setPassed(false);
	      cout << "Wrong!" << endl
		   << "The Correct answer was: "
		   << vp->getStringAnswers(backw)
		   << endl << endl;
	    }
	  else
	    cout << "Wrong!\t" << setting_nrOfRetrys - j
		 << " try(s) left " << endl << endl;
	}
    }
  return retval;
}

// Test a voc... vocNr & nrOfVocs is for displaying (vocNr / nrOfVocs)
string testVoc(VocPair* vp, const bool& backw,
	       const unsigned int& vocNr, const unsigned int& nrOfVocs)
{
  cout << vp->getQuestion(backw) << "\t\t (" << vocNr << "/"
       << nrOfVocs << ")" << endl;

  string answer = readline("");
  rmLeadWs(answer);
  rmTrailWs(answer);
  return answer;
}


// Checks a list for the answer. Will delete the correct anser from the list.
// Returns a boolean representing whether the answer was found or not.
bool isAnsInList(deque<string>& ansList, const string& ans)
{
  unsigned int j = 0;
  while(j < ansList.size())
    {
      if(ans == ansList[j])
	{
	  deque<string>::iterator iter = ansList.begin() + j;
	  ansList.erase(iter);
	  
	  return true;
	}
      else
	j++;
    }
  return false;
}


// If the string (row) is a setting, this will be applyed
// Returns a true if it was a setting and false if it was not
bool ifSettingApply(string& row)
{
  rmComments(row);
  rmLeadWs(row);
  rmTrailWs(row);
      
  unsigned int pos = row.find("=");
  // not (not found) && not last && not first
  if((row.find("=") != string::npos) && (pos != row.length() - 1) && (pos != 0))
    {
      // Filter setting & value
      string setting = row.substr(0, pos);
      string value = row.substr(pos + 1, row.length());
      
      rmTrailWs(setting);
      
      // Go from "arg" to just arg
      pos = value.find("\"");
      value = value.substr(pos + 1, value.length());
      pos = value.find("\"");
      value = value.substr(0, pos);
      
      if(setting == "langDelim")
	{
	  setting_langDelim = value;
	  return true;
	}
      else if(setting == "wordDelim")
	{
	  setting_wordDelim = value;
	  return true;
	}
      else if(setting == "nrOfRetrys")
	{
	  setting_nrOfRetrys = atoi(value.c_str());
	  return true;
	}
      else if(setting == "useVocDir" && (value == "yes" || value == "no"))
	{
	  if(value == "yes")
	    setting_useVocDir = true;
	  else if(value == "no")
	    setting_useVocDir = false;
	  return true;
	}
      else if(setting == "vocDir")
	{
	  fullPathify(value);
	  setting_vocDir = value;
	  return true;
	}
    }
  return false;
}

// Ask the user if he wants to overwrite a certain file.
// Returns true for yes and false for no
bool wantsToOverwriteFile(const string &file)
{
  cout << "Overwrite file " << file << "? (y/n) ";
  
  string input = "";
  while(!(input == "y" || input == "n"))
    {
      input = readline("");
      rmLeadWs(input);
      rmTrailWs(input);
    }
  return input == "y";
}

// Write to the file if you can and display error messages otherwise
void writeIfPossible(string& file, const string& data)
{
  fullPathify(file);

  ifstream inFile(file.c_str());
  bool fileExists = inFile; // True if exists, false otherwise
  inFile.close();

  ofstream outFile(file.c_str());
  if(!outFile)
    cout << "Could not open file " << file << " for reading" << endl;
  else if(!fileExists || (fileExists && wantsToOverwriteFile(file)))
    outFile << data;
  else
    cout << "File NOT saved" << endl;
  
  outFile.close();
}

// Take a file containing ~ and turn it into a full path
// Also takes care of apppending the vocDir path if we're using that
void fullPathify(string& file)
{
  if(file.substr(0, 1) == "~") // The first char in the filename is a ~
    file = ((string) getenv("HOME")) + file.substr(1, file.length());
  else if(setting_useVocDir) // If it wasn't a ~ => see if we've got VocDir enabled
    file = setting_vocDir + file;
}
